#!/usr/bin/python3

# Copyright 2025 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from typing import Optional

IS_WINDOWS = (os.name == "nt")

if IS_WINDOWS:
    import msvcrt
else:
    import termios
    import atexit
    from select import select

class KBHit():
    def __init__(self, listen: bool = False) -> None:
        """Initialize terminal listening for non-blocking keyboard input (noop for windows)"""
        self.is_listening = IS_WINDOWS
        if not IS_WINDOWS:
            # Save the terminal settings
            self.fd = sys.stdin.fileno()
            self.new_term = termios.tcgetattr(self.fd)
            self.old_term = termios.tcgetattr(self.fd)

            # New terminal setting unbuffered
            self.new_term[3] = (self.new_term[3] & ~termios.ICANON & ~termios.ECHO)

            if listen:
                self.listen_term()

            # Support normal-terminal reset at exit
            atexit.register(self.reset_term)

    def __enter__(self):
        """Enter into the listenting terminal setup"""
        self.listen_term()
        return self

    def __exit__(self, exec_type, exec_value, exec_traceback):
        """Exit back into origianl terminal setup"""
        self.reset_term()

    def listening(self) -> bool:
        """Inform program that object is listening for key presses"""
        return self.is_listening

    def reset_term(self) -> None:
        """Reset to original terminal settings"""
        if not IS_WINDOWS and self.is_listening:
            termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.old_term)
            self.is_listening = False

    def listen_term(self) -> None:
        """Switch into listenting terminal setup"""
        if not IS_WINDOWS and not self.is_listening:
            termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.new_term)
            self.is_listening = True

    def getch(self) -> Optional[str]:
        """Read character from standard input if key has been pressed"""
        if self.__kbhit():
            if IS_WINDOWS:
                return msvcrt.getch().decode('utf-8')
            elif self.is_listening:
                return sys.stdin.read(1)

        return None

    def anyKeyPressed(self) -> bool:
        """Check if 'any' key is pressed"""
        hit = self.__kbhit() == True
        if hit and IS_WINDOWS:
            consume = msvcrt.getch()
        return hit

    def __kbhit(self) -> Optional[bool]:
        """Check if key has been pressed"""
        if IS_WINDOWS:
            return msvcrt.kbhit()
        elif self.is_listening:
            dr,dw,de = select([sys.stdin], [], [], 0)
            return dr != []
        else:
            return None



from typing import Optional


#
# In version 1, on Linux and Mac, we explicitly set CONDOR_CONFIG to /dev/null
# if we don't think HTCondor will find it.  I don't know why we don't do this
# on Windows.  In version 1, we forgot to check ~/.condor as well.
#
def _set_null_config():
    import platform as _platform
    import os as _os
    import warnings as _warnings

    if "CONDOR_CONFIG" in _os.environ:
        return

    if _platform.system() in ["Linux", "Darwin"]:
        condor_config_paths = (
            _os.path.expanduser("~/.condor/condor_config"),
            "/etc/condor/condor_config",
            "/usr/local/etc/condor_config",
            _os.path.expanduser("~condor/condor_config"),
        )

        if not any(_os.path.isfile(path) for path in condor_config_paths):
            message = "The environment variable CONDOR_CONFIG is unset and none of the default locations contain a condor_config file.  Using /dev/null, instead."
            _warnings.warn(message)
            _os.environ["CONDOR_CONFIG"] = "/dev/null"


_set_null_config()


#
# Python > 3.8 on Windows doesn't use PATH to find DLLs any more,
# which breaks everything.
#
def _add_dll_dir():
    import platform as _platform

    if _platform.system() in ["Windows"]:
        import sys as _sys
        import os as _os
        from os import path as _path

        if _sys.version_info >= (3,8):
            bin_path = _path.realpath(_path.join(__file__, r'..\..\..\..\bin'))
            return _os.add_dll_directory(bin_path)
    else:
        from contextlib import contextmanager as _contextmanager
        @_contextmanager
        def _noop():
            yield None

        return _noop()


with _add_dll_dir():

    # Modules.
    from ._common_imports import classad

    # Module variables.
    from ._param import _param as param

    # Module functions.
    from .htcondor2_impl import _version as version
    from .htcondor2_impl import _platform as platform
    from .htcondor2_impl import _set_subsystem as set_subsystem
    from .htcondor2_impl import _reload_config_usermaps as reload_config_usermaps
    from .htcondor2_impl import _load_a_config_usermap_file as load_a_config_usermap_file

    from ._security_context import SecurityContext

    from ._loose_functions import send_command
    from ._loose_functions import ping
    from ._loose_functions import send_alive
    from ._loose_functions import set_ready_state
    from ._loose_functions import reload_config
    from ._loose_functions import enable_debug

    from .htcondor2_impl import _disable_debug as disable_debug
    from .htcondor2_impl import _enable_log as enable_log
    from ._logging import _log as log

    # Enumerations.
    from ._subsystem_type import SubsystemType
    from ._daemon_type import DaemonType
    from ._ad_type import AdType
    from ._drain_type import DrainType
    from ._vacate_type import VacateType
    from ._completion_type import CompletionType
    from ._cred_type import CredType
    from ._query_opt import QueryOpt
    from ._job_action import JobAction
    from ._transaction_flag import TransactionFlag
    from ._job_event_type import JobEventType
    from ._file_transfer_event_type import FileTransferEventType
    from ._log_level import LogLevel
    from ._daemon_command import DaemonCommand
    from ._submit_method import SubmitMethod
    from ._history_src import HistorySrc
    from ._common_files_event_type import CommonFilesEventType

    # Classes.
    from ._collector import Collector
    from ._negotiator import Negotiator
    from ._startd import Startd
    from ._placementd import Placementd
    from ._credd import Credd
    from ._cred_check import CredCheck
    from ._schedd import Schedd
    from ._submit import Submit
    from ._submit_result import SubmitResult
    from ._submit_result import _SpooledProcAdList
    from ._job_event import JobEvent
    from ._job_event_log import JobEventLog
    from ._job_status import JobStatus
    from ._remote_param import RemoteParam
    from ._dagman import DAGMan

    # Additional aliases for compatibility with the `htcondor` module.
    from ._daemon_type import DaemonType as DaemonTypes
    from ._ad_type import AdType as AdTypes
    from ._cred_type import CredType as CredTypes
    from ._drain_type import DrainType as DrainTypes
    from ._transaction_flag import TransactionFlag as TransactionFlags
    from ._query_opt import QueryOpt as QueryOpts
    from ._daemon_command import DaemonCommand as DaemonCommands

    # Exceptions.
    from .htcondor2_impl import HTCondorException



#!/usr/bin/env pytest

import os
import htcondor2

class TestHTCondor2Param:

    def test_param(self):
        os.environ['_CONDOR_FOO'] = 'BAR'
        os.environ['_CONDOR_EMPTY'] = '';
        htcondor2.reload_config()

        assert "FOO" in htcondor2.param.keys()
        assert "FOO" in htcondor2.param
        assert htcondor2.param["FOO"] == "BAR"

        assert not "EMPTY" in htcondor2.param.keys()
        assert not "EMPTY" in htcondor2.param

        assert not "OOF" in htcondor2.param.keys()
        assert not "OOF" in htcondor2.param



# Copyright 2022 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random
import pprint
import logging

from pathlib import Path
from operator import itemgetter
from collections import defaultdict

try:
    import elasticsearch
    from elasticsearch import VERSION as ES_VERSION
    _ES_MODULE_FOUND = True
except ModuleNotFoundError as err:
    _ES_MODULE_FOUND = False
    _ES_MODULE_NOT_FOUND_ERROR = err

import adstash.convert as convert
from adstash.utils import get_host_port
from adstash.interfaces.generic import GenericInterface

ES8 = (8,0,0)
if _ES_MODULE_FOUND and (ES_VERSION < (7,0,0) or ES_VERSION >= (9,0,0)):
    logging.warning(f"Unsupported Elasticsearch Python library {ES_VERSION}, proceeding anyway...")


class ElasticsearchInterface(GenericInterface):


    def __init__(
            self,
            host="localhost",
            port="9200",
            url_prefix="",
            username=None,
            password=None,
            use_https=False,
            ca_certs=None,
            timeout=60,
            **kwargs
            ):
        if not _ES_MODULE_FOUND:  # raise module not found error if missing
            raise _ES_MODULE_NOT_FOUND_ERROR
        self.host, self.port = get_host_port(host, port)
        self.url_prefix = url_prefix or ""
        self.username = username
        self.password = password
        self.use_https = use_https
        self.ca_certs = ca_certs
        self.timeout = timeout
        self.handle = None


    def get_handle(self):
        if self.handle is not None:
            return self.handle

        client_options = {}

        if ES_VERSION < ES8:
            client_options["hosts"] = [{
                "host": self.host,
                "port": self.port,
                "url_prefix": self.url_prefix,
                "use_ssl": self.use_https,
                }]
        elif ES_VERSION >= ES8:
            client_options["hosts"] = f"""http{"s" * self.use_https}://{self.host}:{self.port}/{self.url_prefix}"""

        if (self.username is None) and (self.password is None):
            pass  # anonymous auth
        elif (self.username is None) != (self.password is None):
            logging.warning("Only one of username and password have been defined, attempting anonymous connection to Elasticsearch")
        else:  # basic auth
            auth_tuple = (self.username, self.password,)
            if ES_VERSION < ES8:
                client_options["http_auth"] = auth_tuple
            elif ES_VERSION >= ES8:
                client_options["basic_auth"] = auth_tuple

        if self.ca_certs is not None:
            client_options["ca_certs"] = self.ca_certs
        if self.use_https:
            client_options["verify_certs"] = True

        client_options["timeout"] = self.timeout

        self.handle = elasticsearch.Elasticsearch(**client_options)
        return self.handle


    def get_active_index(self, alias):
        client = self.get_handle()
        indices = client.indices.get_alias(name=alias)

        # find which index is reporting as writable
        for index, alias_info in indices.items():
            if alias_info["aliases"][alias].get("is_write_index"):
                logging.info(f"{alias} is an alias, found active index {index}.")
                return index

        # fallback to lexigraphically last index
        indices = list(indices.keys())
        indices.sort(reverse=True)
        logging.warning(f"Could not find an activate index for alias {alias}, trying {indices[0]}")
        return indices[0]


    def get_mappings(self, index):
        client = self.get_handle()
        return client.indices.get_mapping(index=index)[index]["mappings"]


    def make_mappings(self, ad_source=None, **kwargs):
        properties = {}

        properties["metadata"] = {
            "properties": {
                "adstash_runtime": {"type": "date", "format": "epoch_second"},
                "condor_history_runtime": {"type": "date", "format": "epoch_second"},
            },
            "type": "nested",
            "dynamic": True,
        }

        if ad_source is None:

            for name in convert.TEXT_ATTRS:
                properties[name] = {"type": "text"}
            for name in convert.INDEXED_KEYWORD_ATTRS:
                properties[name] = {"type": "keyword"}
            for name in convert.NOINDEX_KEYWORD_ATTRS:
                properties[name] = {"type": "keyword", "index": "false"}
            for name in convert.FLOAT_ATTRS:
                properties[name] = {"type": "double"}
            for name in convert.INT_ATTRS:
                properties[name] = {"type": "long"}
            for name in convert.DATE_ATTRS:
                properties[name] = {"type": "date", "format": "epoch_second"}
            for name in convert.BOOL_ATTRS:
                properties[name] = {"type": "boolean"}
            for name in convert.NESTED_ATTRS:
                properties[name] = {"type": "nested", "dynamic": True}

            dynamic_templates = [
                {
                    "raw_expressions": {  # Attrs ending in "_EXPR" are generated during
                        "match": "*_EXPR",  # ad conversion for expressions that cannot be evaluated
                        "mapping": {"type": "keyword", "index": "false", "ignore_above": 256},
                    }
                },
                {
                    "date_attrs": {  # Attrs ending in "Date" are usually timestamps
                        "match": "*Date",
                        "mapping": {"type": "date", "format": "epoch_second"},
                    }
                },
                {
                    "provisioned_attrs": {  # Attrs ending in "Provisioned" are
                        "match": "*Provisioned",  # resource numbers
                        "mapping": {"type": "long"},
                    }
                },
                {
                    "resource_request_attrs": {  # Attrs starting with "Request" are
                        "match_pattern": "regex",  # usually resource numbers
                        "match": "^Request[A-Z].*$",
                        "mapping": {"type": "long"},
                    }
                },
                {
                    "target_boolean_attrs": {  # Attrs starting with "Want", "Has", or
                        "match_pattern": "regex",  # "Is" are usually boolean checks on the
                        "match": "^(Want|Has|Is)[A-Z_].*$",  # target machine
                        "mapping": {"type": "boolean"},
                    }
                },
                {
                    "strings_as_keywords": {  # Store unknown strings as keywords
                        "match_mapping_type": "string",
                        "mapping": {"type": "keyword", "norms": "false", "ignore_above": 256},
                    }
                },
            ]

        else:
            for name in ad_source.text_attrs:
                properties[name] = {"type": "text"}
            for name in ad_source.indexed_keyword_attrs:
                properties[name] = {"type": "keyword"}
            for name in ad_source.noindex_keyword_attrs:
                properties[name] = {"type": "keyword", "index": "false"}
            for name in ad_source.float_attrs:
                properties[name] = {"type": "double"}
            for name in ad_source.int_attrs:
                properties[name] = {"type": "long"}
            for name in ad_source.date_attrs:
                properties[name] = {"type": "date", "format": "epoch_second"}
            for name in ad_source.bool_attrs:
                properties[name] = {"type": "boolean"}
            for name in ad_source.nested_attrs:
                properties[name] = {"type": "nested", "dynamic": True}
            dynamic_templates = ad_source.dynamic_templates

        mappings = {
            "dynamic_templates": dynamic_templates,
            "properties": properties,
            "date_detection": False,
            "numeric_detection": False,
        }

        return mappings


    def make_settings(self, **kwargs):
        settings = {
            "analysis": {
                "analyzer": {
                    "analyzer_keyword": {"tokenizer": "keyword", "filter": "lowercase"}
                }
            },
            "mapping.total_fields.limit": 5000,
        }
        return settings


    def setup_index(self, index, log_mappings=True, log_dir=Path.cwd(), **kwargs):
        client = self.get_handle()

        # check if index is an alias, and get the active index if so
        if client.indices.exists_alias(name=index):
            index = self.get_active_index(index)

        mappings = self.make_mappings(**kwargs)
        settings = self.make_settings(**kwargs)

        if not client.indices.exists(index=index):  # push new index if doesn't exist
            logging.info(f"Creating new index {index}.")
            client.indices.create(index=index, mappings=mappings, settings=settings)
            if log_mappings and log_dir:
                mappings_file = log_dir / "condor_adstash_elasticsearch_last_mappings.json"
                logging.debug(f"Writing new mappings to {mappings_file}.")
                json.dump(mappings, open(mappings_file, "w"), indent=2)
            return

        # otherwise check existing index for missing mappings properties
        update_mappings = False
        updated_mappings = {}
        existing_mappings = self.get_mappings(index)
        for outer_key in mappings:
            if outer_key not in existing_mappings:  # add anything missing
                updated_mappings[outer_key] = mappings[outer_key]
                update_mappings = True
            elif isinstance(mappings[outer_key], dict):  # update missing keys in any existing dicts
                missing_inner_keys = set(mappings[outer_key]) - set(existing_mappings[outer_key])
                if len(missing_inner_keys) > 0:
                    updated_mappings[outer_key] = {}
                    for inner_key in missing_inner_keys:
                        updated_mappings[outer_key][inner_key] = mappings[outer_key][inner_key]
                    update_mappings = True
        if update_mappings:
            logging.info(f"Updated mappings for index {index}")
            logging.debug(f"{pprint.pformat(updated_mappings)}")
            if ES_VERSION < ES8:
                client.indices.put_mapping(index=index, body=json.dumps(updated_mappings))
            elif ES_VERSION >= ES8:
                client.indices.put_mapping(index=index, **updated_mappings)
            if log_mappings and log_dir:
                mappings_file = log_dir / "condor_adstash_elasticsearch_last_mappings.json"
                logging.debug(f"Writing updated mappings to {mappings_file}.")
                json.dump(updated_mappings, open(mappings_file, "w"), indent=2)


    def make_bulk_body(self, ads, metadata={}):
        body = ""
        for doc_id, ad in ads:
            if metadata:
                ad.setdefault("metadata", {}).update(metadata)
            action = {"index": {"_id": doc_id}}
            body += f"{json.dumps(action)}\n{json.dumps(ad)}\n"
        return body


    def get_error_info(self, result):
        if not result["errors"]:
            return 0

        n_errors = 0
        error_types = defaultdict(int)
        error_reasons = []
        for item in result["items"]:
            try:
                error = item["index"]["error"]
                n_errors += 1
            except (KeyError, TypeError):
                continue
            try:
                error_reasons.append(error["reason"])
            except (KeyError, TypeError):
                pass
            try:
                error_type = error["type"]
            except (KeyError, TypeError):
                error_type = "unknown"
            error_types[error_type] += 1

        error_type_list = list(error_types.items())
        error_type_list.sort(key=itemgetter(1), reverse=True)
        error_type_strs = []
        for (error_type, n) in error_type_list[:3]:
            error_type_strs.append(f"{error_type} ({n} times)")
        logging.error(f"{n_errors} errors encountered during bulk index.")
        logging.error(f"""Most common error type(s): {", ".join(error_type_strs)}.""")
        try:
            logging.error(f"""Example reason: {random.choice(error_reasons)}.""")
        except IndexError:
            pass

        return n_errors


    def post_ads(self, ads, index, metadata={}, **kwargs):
        client = self.get_handle()

        self.setup_index(index, **kwargs)
        body = self.make_bulk_body(ads, metadata)
        result = client.bulk(body=body, index=index)
        n_errors = self.get_error_info(result)
        return {"success": len(ads)-n_errors, "error": n_errors}



#!/usr/bin/env pytest

from ornithology import *
import htcondor2
import os
import re

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
    SHADOW_DEBUG = $(SHADOW_DEBUG) D_TEST D_CAT
    PERIODIC_EXPR_INTERVAL = 1
    SYSTEM_PERIODIC_HOLD = (NumShadowExceptions >= 1)
"""
#endtestreq

NUM_INPUTS = 3000
INPUT_DIR = "input"
OUTPUT_DIR = "output"
ALWAYS_FILE = "always.txt"
TOTAL_BYTES = None

@action
def submit_job(default_condor, test_dir, path_to_sleep):
    global TOTAL_BYTES

    input_list = ALWAYS_FILE
    output_list = ALWAYS_FILE

    if not os.path.exists(ALWAYS_FILE):
        with open(ALWAYS_FILE, "w") as f:
            f.write("This file is always the first transfer item\n")

    if not os.path.exists(INPUT_DIR):
        os.mkdir(INPUT_DIR)

    for i in range(NUM_INPUTS):
        fname = f"omg-why-is-this-filename-so-long-you-may-ask-well-it-is-to-increase-the-number-bytes-passed-in-the-pipe{i:010}.txt"

        input_file = os.path.join(INPUT_DIR, fname)
        if not os.path.exists(input_file):
            with open(input_file, "w") as f:
                f.write("\n")

        input_list += f",{input_file}"
        output_list += f",{fname}"

    TOTAL_BYTES = len(fname) * NUM_INPUTS

    submit = htcondor2.Submit(f"""
        executable = {path_to_sleep}
        arguments  = 0
        log        = job.log

        transfer_input_files = {input_list}

        transfer_output_files = {output_list}
        My.OutputDirectory = "{OUTPUT_DIR}"

        should_transfer_files = YES

        queue
    """)

    return default_condor.submit(submit)

@action
def shadow_log(default_condor):
    return default_condor.shadow_log.open()

class TestFTOPipeFullRead:
    def test_no_shadow_exception(self, submit_job):
        assert submit_job.wait(
            condition=ClusterState.all_complete,
            fail_condition=ClusterState.any_held,
            timeout=300
        )

    def test_check_num_reads(self, shadow_log):
        found_msg = 0
        pattern = re.compile(r"PipeReadFullString\([0-9]+\) Total Reads: [0-9]+")
        for msg in shadow_log.read():
            if re.search(pattern, msg.line):
                analyze = msg.line[msg.line.find("PipeReadFullString"):]
                n_bytes, n_reads = re.findall(r"\d+", analyze)

                if int(n_bytes) >= TOTAL_BYTES:
                    found_msg += 1
                    # This assert assumes the default Linux pipe buffer size,
                    # but on machines with very large amounts of memory, the
                    # pipe buffer size may be substantially larger. The assert
                    # is not critical for this test, so we comment it out.
                    
                    #assert int(n_reads) >= 1

        # We expect the large pipe message twice (input && output transfer)
        assert found_msg == 1



#!/usr/bin/env pytest

import htcondor2

from ornithology import (
    action,
)


TEST_CASES = {
    "case_1": (
"""a,b
1,2
3,4
5,6
""",
        [
            {"a": "1", "b": "2"},
            {"a": "3", "b": "4"},
            {"a": "5", "b": "6"},
        ],
    ),
}


@action(params={name: name for name in TEST_CASES})
def the_test_case(request):
    return (request.param, TEST_CASES[request.param])

@action
def the_test_name(the_test_case):
    return the_test_case[0]

@action
def the_itemdata(the_test_case, test_dir, the_test_name):
    the_table = the_test_case[1][0]
    the_table_file = test_dir / f"{the_test_name}.table"
    the_table_file.write_text(the_table)

    the_text = f"queue FROM TABLE {the_table_file.as_posix()}"
    print(the_text)
    the_submit = htcondor2.Submit(the_text)
    return list(the_submit.itemdata())


@action
def the_expected_itemdata(the_test_case):
    return the_test_case[1][1]


class TestQueueFromTable:

    def test_itemdata(self, the_itemdata, the_expected_itemdata):
        assert the_itemdata == the_expected_itemdata



from typing import Union
from typing import Optional
from typing import List
from typing import Tuple

import socket

from .htcondor2_impl import _handle as handle_t

from .htcondor2_impl import _collector_init
from .htcondor2_impl import _collector_query
from .htcondor2_impl import _collector_locate_local
from .htcondor2_impl import _collector_advertise

from ._ad_type import AdType
from ._daemon_type import DaemonType
from ._security_context import SecurityContext

from ._common_imports import classad

def _ad_type_from_daemon_type(daemon_type: DaemonType):
    map = {
        DaemonType.Master: AdType.Master,
        DaemonType.Startd: AdType.Startd,
        DaemonType.Schedd: AdType.Schedd,
        DaemonType.Negotiator: AdType.Negotiator,
        DaemonType.Generic: AdType.Generic,
        DaemonType.HAD: AdType.HAD,
        DaemonType.Credd: AdType.Credd,
        DaemonType.Collector: AdType.Collector,
        DaemonType.Placementd: AdType.Placementd,
    }
    # Should raise HTCondorEnumError.
    return map.get(daemon_type, None)


class Collector():
    """
    The collector client.  Locate one or more daemons, query for arbitrary
    ads, ask arbitrary daemons directly for their ads, or push new ads into
    the collector.
    """

    #
    # In version 1, there was a distinct DaemonLocation type (a named tuple)
    # that `pool` could also be, but that functionality was never documented.
    #
    def __init__(self, pool : Union[str, classad.ClassAd, List[str], Tuple[str, ...], None] = None, security : Optional[SecurityContext] = None ):
        """
        :param pool:  A ``host::port`` string specifying the remote collector,
                      a list (or tuple) of such strings, or a ClassAd
                      with a ``MyAddress`` attribute (such as might be returned
                      by :meth:`locate`).  :py:obj:`None` means the value of the
                      configuration parameter ``COLLECTOR_HOST``.
        :param security: SecurityContext to use for authentication.
        """
        self._handle = handle_t()

        token = None
        if security is not None:
            token = security.preferred_token

        if pool is None or isinstance(pool, str):
            _collector_init(self, self._handle, pool, token)
            return

        if isinstance(pool, classad.ClassAd):
            addr = pool.get("MyAddress")
            if addr is None:
                raise ValueError("if ClassAd, pool must have MyAddress attribute")
            _collector_init(self, self._handle, addr, token)
            return

        if isinstance(pool, (list, tuple)):
            # For now, just assume that the elements are strings.
            str_list = ", ".join(pool)
            _collector_init(self, self._handle, str_list, token)
            return

        raise TypeError("pool is not a string, list (or tuple) of strings, or a ClassAd")


    # In version 1, `constraint` could also be an ExprTree.  It wouldn't
    # be hard to support, but since we don't want you creating ExprTrees
    # in Python, we're deprecating it and seeing it anyone notices.
    def query(self,
      ad_type: AdType = AdType.Any,
      constraint: Optional[str] = None,
      projection: Optional[List[str]] = None,
      # The documentation does not match the implementation in version 1.
      # statistics: Optional[List[str]] = None,
      statistics: Optional[str] = None,
    ) -> List[classad.ClassAd]:
        r"""
        Returns :class:`classad2.ClassAd`\(s) from the collector.

        :param ad_type:  The type of ClassAd to return.
        :param constraint:  Only ads matching this expression are returned;
                            if not specified, all ads are returned.
        :param projection:  Returned ClassAds will have only these
                            attributes (plus a few the collector insists on).
                            If not specified, return all attributes of each ad.
        :param statistics:  Specify the additional statistical attributes to
                            include, if available, using the same syntax as
                            `STATISTICS_TO_PUBLISH <https://htcondor.readthedocs.io/en/latest/admin-manual/configuration-macros.html#STATISTICS_TO_PUBLISH>`_.
        """
        # str(None) is "None", which is a valid ClassAd expression (a bare
        # attribute reference), so convert to the empty string, instead.
        # We don't pass `constraint` through unmodified because we'll want
        # the str() conversions for all of the other data types we care
        # about to work anyway, and it's easier to do/handle the conversion
        # in Python.
        if constraint is None:
            constraint = ""
        if projection is None:
            projection = []

        # We should consider converting automagically, since it's absurd
        # that this is the only function in the class which takes an AdType
        # instead of a DaemonType.
        if not isinstance(ad_type, AdType):
            raise TypeError("ad_type must be an htcondor.AdType");

        return _collector_query(self._handle, int(ad_type), str(constraint), projection, statistics, None)


    def directQuery(self,
      daemon_type: DaemonType,
      name: Optional[str] = None,
      projection: Optional[List[str]] = None,
      # The documentation does not match the implementation in version 1.
      # statistics: Optional[List[str]] = None,
      statistics: str = None,
    ) -> classad.ClassAd:
        """
        As :meth:`query`, except querying the specified type of daemon rather
        than the collector, which means that at most one ad will be returned.

        :param daemon_type:  Which type of daemon to query.  Some daemons
                             respond to more than one daemon type.
        :param name:  The daemon's name (the ``Name`` attribute from its
                      ad in the collector).  If :py:obj:`None`, use the
                      local daemon.
        :param projection:  See :meth:`query`.
        :param statistics:  See :meth:`query`.
        """
        daemon_ad = self.locate(daemon_type, name)
        daemon = Collector(daemon_ad)
        ad_type = _ad_type_from_daemon_type(daemon_type)
        return daemon.query(ad_type, None, projection, statistics)[0]


    _for_location = ["MyAddress", "AddressV1", "CondorVersion", "CondorPlatform", "Name", "Machine"]


    def locate(self,
        daemon_type: DaemonType,
        name: Optional[str] = None,
    ) -> classad.ClassAd:
        """
        Return a ClassAd with enough data to contact a
        daemon (of the specified type) known to the collector.

        If more than one daemon matches, return only the first.  Use
        :meth:`locateAll` to find all daemons of a given type, or
        :meth:`query` for all daemons of a given name.

        :param daemon_type:  The type of daemon to locate.
        :param name:  The daemon's name (the ``Name`` attribute from its
                      ad in the collector).  If :py:obj:`None`, look for
                      local daemon of the specified type.
        """
        ad_type = _ad_type_from_daemon_type(daemon_type)
        if name is not None:
            constraint = f'Name == "{name}"'

            # Daemon::locate() calls getDaemonInfo(), which has special
            # code for DT_STARTD: if name does not contain an @, figure
            # out the full hostname and match that against the `Machine`
            # attribute, instead.
            if daemon_type == DaemonType.Startd and "@" not in name:
                # This is, after tracing all the scattered logic, how
                # Daemon::locate() converts a hostname into an FQDN.
                r = socket.getaddrinfo(
                        name, None,
                        socket.AF_UNSPEC, socket.SOCK_STREAM,
                        socket.IPPROTO_TCP, socket.AI_CANONNAME
                )
                if len(r) == 0:
                    return None
                constraint = f'Machine == "{r[0][3]}"'

            list = _collector_query(self._handle, int(ad_type), constraint, Collector._for_location, None, name)
            if len(list) == 0:
                return None
            return list[0]

        if self.default is False:
            constraint = True
            list = self.query(ad_type, constraint, Collector._for_location)
            if len(list) == 0:
                return None
            return list[0]

        return _collector_locate_local(self, self._handle, int(daemon_type))


    def locateAll(self,
        daemon_type: DaemonType,
    ) -> List[classad.ClassAd]:
        r"""
        Return a list of :class:`classad2.ClassAd`\s, each enough data to
        contact a daemon known of the given type to the collector.

        :param daemon_type:  Which type of daemons to locate.
        """
        ad_type = _ad_type_from_daemon_type(daemon_type)
        return self.query(ad_type, projection=Collector._for_location)


    def advertise(self,
        ad_list : List[classad.ClassAd],
        command: str = "UPDATE_AD_GENERIC",
        use_tcp: bool = True,
    ) -> None:
        r'''
        Add ClassAd(s) to the collector.

        :param ad_list:  The ClassAd(s) to advertise.
        :param command:  The "advertise command" specifies which kind of
                         ad is being added to the collector.  Different
                         types of ads require different authorization.
                         The list of valid "advertising commands" is kept
                         in the `condor_advertise manpage <https://htcondor.readthedocs.io/en/latest/man-pages/condor_advertise.html>`_.
        :param use_tcp:  Never set this to :py:obj:`False`.
        '''
        _collector_advertise(self._handle, ad_list, command, use_tcp)



#!/usr/bin/env pytest


from ornithology import *
import htcondor2 as htcondor
import os

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
    # Each job list contains 7 jobs so tolerate 6 failures and
    # fail batch on 7th failure to ensure all jobs run but list fails
    DAGMAN_NODE_JOB_FAILURE_TOLERANCE = 6
"""
#endtestreq

@action
def writeExecutable(test_dir):
    file = test_dir / "exit.sh"
    with open(file, "w") as f:
        f.write("""#!/bin/bash\nexit $1\n""")
    os.chmod(str(file), 0o0777)
    return file

@action
def writeItemData(test_dir):
    good_data = test_dir / "good.data"
    with open(good_data, "w") as f:
        f.write("""0\n0\n0\n0\n""")
    bad_data = test_dir / "bad.data"
    with open(bad_data, "w") as f:
        f.write("""1\n1\n2\n3\n10\n10\n10\n""")
    return (good_data, bad_data)

@action
def writeSubmitFile(test_dir, writeExecutable):
    file = test_dir / "job.sub"
    with open(file, "w") as f:
        f.write(f"""
executable = {writeExecutable}
arguments  = "$(code)"
log        = jobs.log
priority   = $(code) * 100

periodic_remove = $(DO_REMOVE) && $(ProcId) >= 4
requirements    = !$(DO_REMOVE) || $(ProcId) < 4

queue code from $(filename)
        """)
    return file

@action
def writePostScript(test_dir):
    file = test_dir / "check-macros.py"
    with open(file, "w") as f:
        f.write("""#!/usr/bin/env python3
import sys
import htcondor2 as htcondor

if len(sys.argv) != 22:
    print(f"Unexepected number of arguments: {len(sys.argv)}")
    sys.exit(1)

schedd = htcondor.Schedd()
ads = schedd.query(constraint=f"DAG_NodesTotal=!=UNDEFINED", projection=['ClusterId'])

if len(ads) > 1:
    print("Schedd Query returned more than one ad (expected only the DAGMan job)")
    for ad in ads:
        print(ad)
    sys.exit(1)

DAG_ID = ads[0]["ClusterId"]

ads = schedd.history(constraint=f"DAGManJobId=={DAG_ID}", projection=["ClusterId", "DAGNodeName"], match=1)

CLUSTER_ID = 0
NODE_NAME = "???"
for ad in ads:
    CLUSTER_ID = ad["ClusterId"]
    NODE_NAME = ad["DAGNodeName"]

EXPECTED = {
    "GOOD" : [
        "0", # Retry
        "1", # Max retries
        "0", # DAG status
        "0", # Failed count
        "0", # Futile count
        "0", # Done count
        "0", # Queued count
        "2", # Num nodes
        f"{DAG_ID}", # DAGMan job id
        f"{CLUSTER_ID}.3", # Job ID
        f"{CLUSTER_ID}", # Cluster id
        "4", # Num jobs
        "True", # Success
        "0", # Return
        "0", # Exit codes
        "0:4", # Exit frequencies
        "0,0,0,0", # Exit code list
        "-1", # Pre Script return
        "0", # Num jobs aborted
    ],
    "BAD" : [
        "0", # Retry
        "0", # Max retries
        "0", # DAG status
        "0", # Failed count
        "0", # Futile count
        "1", # Done count
        "0", # Queued count
        "2", # Num nodes
        f"{DAG_ID}", # DAGMan job id
        {f"{CLUSTER_ID}.3", f"{CLUSTER_ID}.4", f"{CLUSTER_ID}.5", f"{CLUSTER_ID}.6"}, # Job ID
        f"{CLUSTER_ID}", # Cluster id
        "7", # Num jobs
        "False", # Success
        {"3","2","1"}, # Return
        "1,2,3", # Exit codes
        "1:2,2:1,3:1", # Exit frequencies
        "1,1,2,3,,,", # Exit code list
        "-1", # Pre Script return
        "3", # Num jobs aborted
    ],
}

if NODE_NAME not in EXPECTED.keys():
    print(f"Unexpected node name '{NODE_NAME}'")
    sys.exit(1)

if NODE_NAME != sys.argv[1] or NODE_NAME != sys.argv[2]:
    print(f"Provided Script $NODE ({sys.argv[1]}) or $JOB ({sys.argv[2]}) does not match {NODE_NAME}")
    sys.exit(1)

failed = False
for i, script_macro in enumerate(sys.argv[3:]):
    expected_val = EXPECTED[NODE_NAME][i]
    match = script_macro in expected_val if type(expected_val) is set else script_macro == expected_val
    if not match:
        print(f"EXPECTED[{NODE_NAME}][{i}] ({expected_val}) != {script_macro}")
        failed = True

if failed:
    sys.exit(1)

print("Script arguments were expected values")

        """)
    os.chmod(str(file), 0o0777)
    return file

@action
def runDAG(default_condor, test_dir, writePostScript, writeSubmitFile, writeItemData):
    DAG_FILE = "test.dag"
    with open(DAG_FILE, "w") as f:
        f.write(f"""
JOB GOOD {writeSubmitFile}
JOB BAD {writeSubmitFile}

PARENT GOOD CHILD BAD

VARS GOOD PREPEND filename="{writeItemData[0]}" DO_REMOVE="False"
VARS BAD PREPEND filename="{writeItemData[1]}" DO_REMOVE="True"

RETRY GOOD 1

SCRIPT DEBUG script.debug ALL POST GOOD {writePostScript} $node $JOB $RETRY $MAX_RETRIES $DAG_STATUS $FAILED_COUNT $FUTILE_COUNT $DONE_COUNT $QUEUED_COUNT $NODE_COUNT $DAGID $JOBID $CLUSTERID $JOB_COUNT $SUCCESS $RETURN $EXIT_CODES $EXIT_CODE_COUNTS $EXIT_CODE_LIST $PRE_SCRIPT_RETURN $JOB_ABORT_COUNT
SCRIPT DEBUG script.debug ALL POST BAD {writePostScript} $NoDe $JOB $RETRY $MAX_RETRIES $DAG_STATUS $FAILED_COUNT $FUTILE_COUNT $DONE_COUNT $QUEUED_COUNT $NODE_COUNT $DAGID $JOBID $CLUSTERID $JOB_COUNT $SUCCESS $RETURN $EXIT_CODES $EXIT_CODE_COUNTS $EXIT_CODE_LIST $PRE_SCRIPT_RETURN $JOB_ABORT_COUNT
""")

    assert os.path.exists(DAG_FILE)
    dag = htcondor.Submit.from_dag(str(DAG_FILE))
    return default_condor.submit(dag)


@action
def checkDAGExit(default_condor, runDAG):
    with default_condor.use_config():
        ads = htcondor.Schedd().history(f"ClusterId=={runDAG.clusterid}", projection=["ExitCode"], match=1)
        for ad in ads:
            return int(ad["ExitCode"])

class TestDAGManScriptMacros:
    def test_dagman_script_macros(self, runDAG):
        # POST scripts verify everything. As long as the DAG completes successfully then pass
        assert runDAG.wait(condition=ClusterState.all_complete, timeout=120)
        assert runDAG.state.all_status(jobs.JobStatus.COMPLETED)

    def test_verify_exit_code(self, checkDAGExit):
        assert checkDAGExit == 0




from flask import Blueprint, render_template,redirect, url_for, request, current_app
import csv
from io import StringIO
import typing as t
import urllib.parse
import urllib.request
import xml.etree.ElementTree as ET
import classad2 as classad
import htcondor2 as htcondor
from dataclasses import dataclass, fields, astuple
from math import floor
import time
import threading
import re
from utils import cache_response_to_disk, make_data_response, getOrganizationFromInstitutionID
# from . import overview  # Import the overview module

##########################################
# Functions to generate the data by querying Topology and Collectors
##########################################

def get_resources_tree(params: str) -> ET.Element:
    """
    Query the Topology resource group XML endpoint with the given parameters.
    Returns an ElementTree Element.

    May raise urllib.error.HTTPError.
    """
    handle = urllib.request.urlopen(
        "https://topology.opensciencegrid.org/rgsummary/xml?%s" % params
    )
    result_bytes = handle.read()

    return ET.fromstring(
        result_bytes.decode(encoding="utf-8", errors="surrogateescape")
    )


def elem_text(elem: ET.Element, path: t.Union[None, str] = None) -> str:
    """
    Get the text from an element (or subelement if `path` is provided), stripped of
    leading and trailing whitespace.
    Return the empty string if the element doesn't exist.
    """
    if path:
        elem = elem.find(path)
    try:
        return elem.text.strip()
    except AttributeError:  # we got a None
        return ""


def is_hosted_fqdn(fqdn):
    """
    Return True if the FQDN is that of a hosted CE

    Match domains that start with hosted-ce and end with any of:
    - grid.uchicago.edu
    - opensciencegrid.org
    - osg-htc.org
    Or domains that end with:
    - svc.opensciencegrid.org
    - svc.osg-htc.org
    """
    return bool(re.match(r"^hosted-ce.*\.(grid\.uchicago\.edu|opensciencegrid\.org|osg-htc\.org)|.*\.svc\.(opensciencegrid|osg-htc)\.org$", fqdn))

@dataclass
class ResourceInfo:
    fqdn: str
    active: bool
    registered: bool = True
    facility_name: str = "Unknown" # a.k.a. institution
    site_name: str = "Unknown"
    description: str = "Unknown"
    name: str = "Unknown" 
    isCCStar: bool = False  # True if this is a CCstar site
    hosted: bool = False
    scheduler: str = "Unknown"
    health: str = "Poor"
    healthInfo: str = "Not reporting; likely not running"
    startTime: str = "Unknown"
    version: str = "Unknown"
    allocationsPastWeek: bool = False
    allocationsPastMonth: bool = False
    glideinsRunning: int = 0
    glideinsIdle: int = 0
    glideinsHeld: int = 0
    def __post_init__(self):
        self.hosted = is_hosted_fqdn(self.fqdn)


@dataclass
class HeaderLink:
    """ Util class containing the text and URL for a header link """
    title: str
    tooltip: str
    url: str

def returnOrAddUnregisteredInfo(resource_info_by_fqdn, fqdn):
    """
    Add an unregistered CE to the resource info dictionary.
    """
    if fqdn not in resource_info_by_fqdn:
        # The CE is not registered in Topology
        resource_info = ResourceInfo(
            fqdn=fqdn,
            active=True,
            registered=False
        )
        resource_info_by_fqdn[fqdn] = resource_info
    return resource_info_by_fqdn[fqdn]

def ce_info_from_ganglia(resource_info_by_fqdn):
    """
    Given a dictionary of CE info from Topology, augment this with information
    gleaned from the time series database.
    """
    import pandas as pd
    host = current_app.config['CE_DASHBOARD_DEFAULT_CE_DOMAIN']
    r = 'month'
    df=pd.read_csv('https://display.ospool.osg-htc.org/ganglia/graph.php?r=' + r + '&hreg[]=(svc.osg-htc.org|' + host + ')&mreg[]=%5E' + 'CpusInUse' + '&aggregate=1&csv=1',skipfooter=1,engine='python')
    
    # Rename 'Timestamp' column to 'Date' for clarity and set it as the index
    df.rename({'Timestamp':'Date'}, axis='columns', inplace=True)
    df.set_index('Date', inplace=True)
    # Convert the index to datetime
    df.index = pd.to_datetime(df.index, utc=True)
    df.index = df.index.tz_localize(None)
    
    # The columns are the FQDNs of the CEs.  The values are the number of CPUs in use at that time.
    # We want a list of the FQDNs that did not use any CPUs at all in the last week, and 
    # a second list of the FQDNs that did not use any CPUs at all in the last month.
    from datetime import datetime, timedelta

    # Define the time ranges for the last week and last month
    one_week_ago = datetime.now() - timedelta(weeks=1)
    one_month_ago = datetime.now() - timedelta(weeks=4)

    # Filter FQDNs that did not use any CPUs in the last week
    fqdn_some_cpus_past_week = [
        fqdn.strip() for fqdn in df.columns
        if df.loc[df.index >= one_week_ago, fqdn].sum() > 0
    ]
    # Set the allocationsPastWeek flag for these FQDNs
    for fqdn in fqdn_some_cpus_past_week:
        info = returnOrAddUnregisteredInfo(resource_info_by_fqdn, fqdn)
        info.allocationsPastWeek = True

    # Filter FQDNs that did not use any CPUs in the last month
    fqdn_some_cpus_past_month = [
        fqdn.strip() for fqdn in df.columns
        if df.loc[df.index >= one_month_ago, fqdn].sum() > 0
    ]
    # Set the allocationsPastMonth flag for these FQDNs
    for fqdn in fqdn_some_cpus_past_month:
        info = returnOrAddUnregisteredInfo(resource_info_by_fqdn, fqdn)
        info.allocationsPastMonth = True

def ce_info_from_collectors(resource_info_by_fqdn):
    """
    Given a dictionary of CE info from Topology, augment this with information
    gleaned from the top-level CE collector, and from querying CE collectors
    for hosted/active CEs.
    """
    # First grab all ads from top level collector
    coll = htcondor.Collector("collector.opensciencegrid.org:9619")
    try:
        ads = coll.query(htcondor.AdTypes.Schedd,projection=["Name","CollectorHost","OSG_BatchSystems","Status","DaemonStartTime","HTCondorCEVersion","TotalRunningJobs","TotalHeldJobs","TotalIdleJobs"])
    except Exception as e:
        print(f"Unable to reach collector.opensciencegrid.org:9619: {e}")
        for fqdn in resource_info_by_fqdn:
            resource_info_by_fqdn[fqdn].health="Unknown"
            resource_info_by_fqdn[fqdn].healthInfo="Server collector.opensciencegrid.org unreachable"
        return
    htcondor.param["QUERY_TIMEOUT"] = "5"
    for ad in ads:
        fqdn = ad["Name"]
        info = returnOrAddUnregisteredInfo(resource_info_by_fqdn, fqdn)
        info.scheduler = ad.get("OSG_BatchSystems","Unknown")
        info.scheduler = info.scheduler[0].upper() + info.scheduler[1:]  # ensure first letter capitalized
        if info.scheduler == "Condor":
            info.scheduler = "HTCondor"
        info.version = ad.get("HTCondorCEVersion","Unknown")
        if "DaemonStartTime" in ad:
            info.startTime = classad.ExprTree('formatTime(DaemonStartTime,"%Y-%m-%d")').eval(ad)
        # If CE is hosted and active, try to query the CE collector for grid ads
        if info.hosted and info.active:
            canReachCeCollector = False
            if "CollectorHost" in ad:
                ceColl = htcondor.Collector(ad["CollectorHost"])
                try:
                    gridAds = ceColl.query(htcondor.AdTypes.Grid,projection=["GridResourceUnavailableTime","GridResourceUnavailableTimeReason"])
                    canReachCeCollector = True
                except:
                    canReachCeCollector = False
            if not canReachCeCollector:
                info.health="Poor"
                info.healthInfo="CE Collector unreachable"
                continue
            for gridAd in gridAds:
                if "GridResourceUnavailableTime" in gridAd and "GridResourceUnavailableTimeReason" in gridAd:
                    hinfo = classad.ExprTree('strcat("Since ",formatTime(GridResourceUnavailableTime),": ",GridResourceUnavailableTimeReason)').eval(gridAd)
                    info.health="Poor"
                    info.healthInfo=hinfo
                    continue
        
        # Check for too many held jobs, or idle jobs without anything running, etc.
        totalHeld = ad["TotalHeldJobs"]
        totalIdle = ad["TotalIdleJobs"]
        totalDaysNoStartedJobs = 0
        if info.allocationsPastWeek == False:
            totalDaysNoStartedJobs = 7
        if info.allocationsPastMonth == False:
            totalDaysNoStartedJobs = 30
        totalRunning  = ad["TotalRunningJobs"]
        if info.hosted or info.scheduler.casefold() != "Condor".casefold():
            # All CEs that are (a) hosted or (b) in front of any system other than
            # HTCondor will have two running jobs per running glidein: the vanilla
            # universe provisioning request, and the routed grid universe.  So
            # for these systems, divide totalRunning by two.
            totalRunning = floor( totalRunning / 2)
        info.glideinsRunning = totalRunning
        info.glideinsIdle = totalIdle
        info.glideinsHeld = totalHeld

        status = "Unknown"
        if "Status" in ad:
            status = str( classad.ExprTree('Status').eval(ad) )

        # Checks here should be done in the order of severity, from most to least 
        if totalRunning == 0 and totalIdle == 0 and totalHeld == 0:
            info.health="Poor"       
            info.healthInfo="No glideins at all at this CE; maybe factory misconfigured?"
            continue
        if totalDaysNoStartedJobs > 29:
            info.health="Poor"       
            info.healthInfo="No allocated Cpus in the past month"
            continue
        if totalDaysNoStartedJobs > 6:
            info.health="Fair"       
            info.healthInfo="No allocated Cpus in the past week"
            continue
        if totalHeld > totalRunning + totalIdle and totalRunning > 0:
            info.health="Fair"       
            info.healthInfo="Many glideins being held: running=" + str(totalRunning) + " held=" + str(totalHeld) + " idle=" + str(totalIdle)
            continue
        # Check for load problems (RecentDaemonCoreDutyCycle and File Transfer load are reflected in Status attr)
        if status == "CRITICAL":
            info.health="Fair"
            info.healthInfo = "Server load level = " + status
            continue
        if status == "WARNING":
            info.health = "Fair"
            info.healthInfo = "Server load level = " + status
            continue
        
        # If we made it here, things look good!
        info.health="Good"
        #info.healthInfo="Glideins running=" + str(totalRunning) + " held=" + str(totalHeld) + " idle=" + str(totalIdle)
        info.healthInfo=""


def ce_info_from_topology() -> t.Dict[str, ResourceInfo]:
    """
    Get the resource info of production CEs from Topology
    """
    params = [
        # "gridtype=on&gridtype_1=on",  # production, not itb
        "service=on&service_1=on",  # resources with CEs
    ]
    tree = get_resources_tree("&".join(params))

    resource_info_by_fqdn = {}

    for resourcegroup in tree.findall("./ResourceGroup"):
        # We want to use the institution ID to get the organization name, as this is the new way to do it.
        # If the institution ID is not found, we will fallback to the old facility name.
        old_facility_name = elem_text(resourcegroup, "./Facility/Name")
        institution_id = elem_text(resourcegroup, "./Facility/InstitutionID")
        facility_name = getOrganizationFromInstitutionID(institution_id, old_facility_name)
        site_name = elem_text(resourcegroup, "./Site/Name")
        for resource in resourcegroup.findall("./Resources/Resource"):
            fqdn = elem_text(resource, "./FQDN")
            if not fqdn:
                continue
            name = elem_text(resource, "./Name")
            description = elem_text(resource, "./Description")
            description = " ".join(description.splitlines())    # Description may have newlines; get rid of em
            description = description.replace('\\','')          # Get rid of escaping backslashes
            description = description.replace('"','')           # Get rid of doublequotes
            description = description.replace(',',';')          # Get rid of commas, as we use commas to delimit fields in the CSV output
            active = elem_text(resource, "./Active").lower()
            isCCStar = elem_text(resource, "./IsCCStar").lower()
            resource_info = ResourceInfo(
                fqdn=fqdn,
                facility_name=facility_name,
                site_name=site_name,
                description=description,
                name=name,
                active=(active == "true"),  # convert string to bool
                isCCStar=(isCCStar == "true"),  # convert string to bool
            )
            resource_info_by_fqdn[fqdn] = resource_info
    return resource_info_by_fqdn

@cache_response_to_disk(file_name="ce_info.csv")
def get_landing_response():
    # Gather up CE info from Topology
    resource_info_by_fqdn = ce_info_from_topology()
    # Now gather up CE info from the ganglia time series database
    ce_info_from_ganglia(resource_info_by_fqdn)
    # Now gather up CE info from the collectors - currently must be done after all other sources
    ce_info_from_collectors(resource_info_by_fqdn)
    # Output this info as a CSV that htcondorview can use
    default_domain = current_app.config['CE_DASHBOARD_DEFAULT_CE_DOMAIN']
    output = StringIO()
    writer = csv.writer(output,quoting=csv.QUOTE_ALL,lineterminator=']')
    output.write('[[')
    writer.writerow([field.name for field in fields(ResourceInfo)])
    for ri in resource_info_by_fqdn.values():
        output.write(',\n[')
        # If the Name is unknown, use the first part of the FQDN
        if ri.name=="Unknown":
            ri.name =  ri.fqdn.split('.')[0]
        # If the FQDN ends with the default domain, remove it
        if ri.fqdn.endswith('.' + default_domain):
            ri.fqdn = ri.fqdn[:-len(default_domain)-1]
        # Add a URL with the name to point to the Overview page if not in Poor health
        if ri.health != "Poor":
            if ri.allocationsPastWeek:
                ri.name=f"/overview.html?host={ri.fqdn}|{ri.name}"
            else:
                # No allocations in the past week, so use the month view
                ri.name=f"/overview.html?host={ri.fqdn}&r=month|{ri.name}"
        writer.writerow(astuple(ri))
    output.write(']')
    result = output.getvalue()
    output.close()
    result = result.replace('"True"','true')
    result = result.replace('"False"','false')
    return result

def update_ce_info():
    """
    Update the in-memory dictionary of CE info
    """
    # check if we updated recently
    global ce_info_dict_last_update_time
    with lock:
        if (time.time() - ce_info_dict_last_update_time) < 60 * 10: # only update if last update was more than 10 minutes ago
            return
        ce_info_dict_last_update_time = time.time()
    # get the response and parse it
    response, _ = get_landing_response()
    global ce_info_dict
    with lock:
        ce_info_dict.clear()
        for line in response.splitlines():
            line = line.rstrip(',')
            if line.startswith('[') and line.endswith(']'):
                values = line.strip('[]').replace('"','').split(',')
                if len(values) == len(fields(ResourceInfo)):
                    ri = ResourceInfo(*values)
                    ce_info_dict[ri.fqdn] = ri

def get_ce_info(fqdn: str) -> t.Optional[ResourceInfo]:
    """
    Get the CE info for a given FQDN from the in-memory dictionary.
    If the FQDN is not found, try updating the dictionary and return None if not found.
    """
    if fqdn in ce_info_dict:
        return ce_info_dict[fqdn]
    # If not found, update the dictionary and try again     
    update_ce_info()
    return ce_info_dict.get(fqdn)

def get_ce_facility_site_descrip(fqdn: str):
    """
    Get the facility/site description for a given FQDN.
    If the FQDN is not found, try updating the dictionary and return None if not found.
    """
    ce_info = get_ce_info(fqdn)
    if ce_info:
        # Re-extract the original site name from the URL if the health isn't Poor
        name = ce_info.name if ce_info.health == "Poor" else ce_info.name.split('|',1)[1]
        return ce_info.facility_name, ce_info.site_name, name, ce_info.description, ce_info.health
    return "Unknown", "Unknown", "Unknown", "Unknown", "Poor"

def get_next_prev_sites(fqdn: str) -> t.Tuple[t.Optional[HeaderLink], t.Optional[HeaderLink]]:
    """
    Given a site FQDN, return the next and previous healthy sites in alphabetical order.
    """
    # Get all CE Facility names in sorted order, for the facilities that have links from the homepage
    healthy_facilities_by_name = sorted(
        (ce for ce in ce_info_dict.values() if 'overview.html' in ce.name and ce.health != "Poor"), 
        key=lambda x: x.facility_name)
    current_index = next((i for i, ri in enumerate(healthy_facilities_by_name) if ri.fqdn == fqdn), 0)

    prev_index = current_index - 1 if current_index > 0  else len(healthy_facilities_by_name) - 1
    next_index = current_index + 1 if current_index < len(healthy_facilities_by_name) - 1 else 0

    prev_facility = healthy_facilities_by_name[prev_index]
    next_facility = healthy_facilities_by_name[next_index]

    # TODO re-extracting the original name and link from the parsed facility csv here is a bit hacky
    prev_link, prev_name = prev_facility.name.split('|',1)
    next_link, next_name = next_facility.name.split('|',1)

    return (
        HeaderLink(title='Next\nCE', tooltip=f"{next_facility.facility_name} - {next_name}", url=next_link),
        HeaderLink(title='Prev\nCE', tooltip=f"{prev_facility.facility_name} - {prev_name}", url=prev_link)
    )

##########################################
# Flask routes
##########################################

landing_bp = Blueprint('landing', __name__)
lock = threading.Lock()
ce_info_dict : dict[str, ResourceInfo] = {}
ce_info_dict_last_update_time = 0

landing_linkmap = {
}

@landing_bp.route('/data/ce_landing')
def ce_landing_data():
    response_body, cached_response = get_landing_response()
    return make_data_response(response_body,cached_response)


@landing_bp.route('/landing.html')
def ce_admin_landing_page():
    return render_template('landing.html',linkmap=landing_linkmap,page_title="Hosted CE Dashboards")

@landing_bp.route('/home.html')
@landing_bp.route('/select.html')
@landing_bp.route('/index.html')
def ce_user_landing_page():
    return render_template('home.html',linkmap={},page_title="Available CE Dashboards")

@landing_bp.route('/')
def ce_goto_default_or_user_landing_page():
    """
    Examine the cookie named ceDashStoredHost to see if the user has a preferred page.
    If so, redirect to that page.
    If not, redirect to the default landing page.
    If the cookie is not found, redirect to the default landing page.       
    """
    stored_page_url = request.cookies.get('ceDashStoredHost')
    if stored_page_url and stored_page_url != "none":
        return redirect(url_for('overview.overview') + '?' + urllib.parse.urlencode({'host': stored_page_url}))
    else:
        # Redirect to the default user landing page
        return redirect(url_for('landing.ce_user_landing_page'))



import os
import re
import pathlib

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

# Nice variable references for parts of
# a version tuple (MAJOR, MINOR, RELEASE) ex: (24.3.1)
MAJOR = 0
MINOR = 1
RELEASE = 2

# Supported history entry types
HISTORY_TYPES = ["bug", "bugs", "feature", "features"]

class FlattenVersionHistory(Directive):
    """Custom directive to read mega version history files and flatten into one list"""

    # Directive setup
    require_arguments = 1 # Requires an entry type (see HISTORY_TYPES)
    optional_arguments = 10 # Up to 10 versions to parse from
    has_content = True # Allow additional content below decalration
    # Extra options
    option_spec = {
        # List of ticket numbers to exclude in this inclusion
        "exclude" : directives.positive_int_list,
    }

    def _excluded_entry(self, entry: list[tuple[str,int]]) -> bool:
        """Check if parsed version history entry is attributed to excluded tickets"""
        # Check if we are even excluding any tickets
        exclude_list = self.options.get("exclude")
        if exclude_list is None or len(exclude_list) == 0:
            return False

        # Resolve entry into a single string
        entry_str = "\n".join([pair[0] for pair in entry])
        # Count the number of specified jira tickets
        num_tickets = entry_str.count(":jira:")

        # Check each specified exclude ticket number for existance in entry
        match_count = 0
        for ticket in set(exclude_list):
            check = f":jira:`{ticket}`"
            if check in entry_str:
                match_count += 1
            if match_count == num_tickets:
                break

        return match_count == num_tickets

    def run(self) -> list[nodes.Node]:
        """Execute this directive"""
        # Used to parse read version history entries into node
        rst = ViewList()
        # Make paragraph node (end result)
        node = nodes.paragraph()

        # Add any additional contents to the node
        rst.extend(self.content)

        # Get absolute path to version-history directory (where history entry files should live)
        vers_hist_dir = pathlib.Path(self.state.document.settings.env.srcdir) / "version-history"

        # Check if required entry type is valid
        ref_type = self.arguments[0].lower()
        if ref_type not in HISTORY_TYPES:
            raise self.error(f"{self.name}: Invalid version history type '{ref_type}'")
        # Check if there have been specified versions to include/flatten
        elif len(self.arguments) > 1:
            # Dictionary of specified versions {MAJOR : set(full versions)}
            specified_versions = dict()

            # Check each argument is valid and parse into specified versions dict for later
            for arg in self.arguments[1:]:
                if re.match(r"(\d+)\.(\d+)\.(\d+)", arg):
                    version = tuple(int(x) for x in arg.split("."))
                    key = version[MAJOR]
                    if key in specified_versions:
                        specified_versions[key].add(version)
                    else:
                        specified_versions[key] = set([version])
                else:
                    raise self.error(f"{self.name}: Invalid version number ({arg}) provided")

            # Reach each associated major version history entry file (v24-version.hist)
            for major, versions in specified_versions.items():
                version_file = vers_hist_dir / f"v{major}-version.hist"
                # Ignore versions if entry file DNE
                if version_file.exists():
                    # !!!! IMPORTANT !!!! - This informs sphinx to rebuild version histories when entry files are updated
                    self.state.document.settings.env.note_dependency(str(version_file))
                    # Read version history entry file
                    with version_file.open() as f:
                        included_entries = [] # Entries to actually include
                        entry = [] # Current entry being parsed (for exclude checking)
                        found_section = False
                        line_no = 0

                        for line in f:
                            if line.strip() == "" or line.strip()[0] == "#" or line.strip()[:2] == "//":
                                # Ignore blank lines & allow '#' and '//' to denote comments
                                pass

                            elif line.strip()[:3] == "***":
                                # Parse section banner line (*** version type)
                                # Everything below this line belongs to this version type

                                # Break if a new section but we are no longer including from this file
                                if len(versions) == 0:
                                    break

                                found_section = False

                                # Attempt to read section banner (*** version type)
                                try:
                                    _, vers, section_type = line.strip().split()[:3]
                                    section_version = tuple(int(x) for x in vers.split("."))

                                    # Make sure specified type is valid
                                    if section_type.lower() not in HISTORY_TYPES:
                                        raise RuntimeError(f"Unknown version history type '{section_type}'")

                                    # Check that directive type matches section type
                                    match_type = (
                                        (ref_type in HISTORY_TYPES[:2] and section_type in HISTORY_TYPES[:2])
                                        or
                                        (ref_type in HISTORY_TYPES[2:] and section_type in HISTORY_TYPES[2:])
                                    )

                                    # Check if section version was specified and type matches (remove from needed versions)
                                    found_section = (section_version in versions and match_type)
                                    if found_section:
                                        versions.remove(section_version)

                                except Exception as e:
                                    raise self.error(f"{self.name}: Invalid version history file '{version_file}': {e}")

                            elif found_section:
                                # Store section lines for later parsing
                                if len(line) > 1 and line[0] == '-':
                                    # We expect all entries to begin w/ '-' with no padding
                                    if len(entry) > 0:
                                        # Already have an entry stored... add to final list if not excluded and clear
                                        if not self._excluded_entry(entry):
                                            included_entries.extend(entry)
                                        entry.clear()
                                # Store line in temporary entry
                                entry.append((line, line_no))
                            line_no += 1

                        # Make sure to include any final entry information if not excluded
                        if len(entry) > 0 and not self._excluded_entry(entry):
                            included_entries.extend(entry)

                        # Add all non-excluded entries to rst parser object
                        for line, num in included_entries:
                            rst.append(line, str(version_file), num)

        # If no information has been parsed (from files or inline content) then default to '- None'
        if len(rst) == 0:
            rst.append("- None.", "Internal-Parse", self.lineno)

        # Parse list into paragraph node
        nested_parse_with_titles(self.state, rst, node)

        return [node]

def setup(app):
    """Setup sphinx to know of this directive"""
    app.add_directive('include-history',  FlattenVersionHistory)




import random
import time
import urllib

import urllib3

from credmon.CredentialMonitors.LocalCredmon import LocalCredmon, TokenInfo

class ClientCredmon(LocalCredmon):
    """
    A CredMon class that uses the OAuth2 Client Credentials flow for
    acquiring access tokens
    """

    @property
    def credmon_name(self):
        return "CLIENT"

    def __init__(self, *args, **kwargs):
        super(ClientCredmon, self).__init__(*args, **kwargs)
        self._token_endpoint = ""
        self._token_endpoint_resolved = 0
        self._token_endpoint_lifetime = int(self.get_credmon_config("TOKEN_ENDPOINT_LIFETIME", "3600")) + random.randint(0, 60)

        self.client_id = self.get_credmon_config("CLIENT_ID")
        if not self.client_id:
            raise Exception(f"{self.credmon_name}_CREDMON_{self.provider}_CLIENT_ID configuration parameter must be set to use the client credential credmon")
        secret_file = self.get_credmon_config("CLIENT_SECRET_FILE")
        if not secret_file:
            raise Exception(f"{self.credmon_name}_CREDMON_{self.provider}_CLIENT_SECRET_FILE configuration parameter must be set to use the client credential credmon")

        with open(secret_file) as fp:
            self.secret = fp.read().strip()

        if not self.secret:
            raise Exception(f"Client credentials flow credmon configured with an empty secret file ({self.credmon_name}_CREDMON_{self.provider}_CLIENT_SECRET_FILE)")


    @property
    def token_endpoint(self):
        """
        Determine the token endpoint from the static configuration or dynamically from
        OIDC metadata discovery
        """
        endpoint = self.get_credmon_config("TOKEN_URL")
        if endpoint:
            return endpoint

        if self._token_endpoint and (self._token_endpoint_resolved + self._token_endpoint_lifetime > time.time()):
            return self._token_endpoint

        resp = urllib3.request("GET", self.token_issuer + "/.well-known/openid-configuration")
        if resp.status != 200:
            msg = f"When performing OIDC metadata discovery, {self.token_issuer} responded with {resp.status}"
            self.log.error(msg)
            if self._token_endpoint:
                return self._token_endpoint
            raise Exception(msg)

        try:
            resp_json = resp.json()
        except Exception as exc:
            self.log.error(f"When performing OIDC metadata discovery, {self.token_issuer} had non-JSON response: {exc}")
            if self._token_endpoint:
                return self._token_endpoint
            raise

        if 'token_endpoint' not in resp_json:
            msg = f"When performing OIDC metadata discovery, {self.token_issuer} response lacked the `token_endpoint` key"
            self.log.error(msg)
            if self._token_endpoint:
                return self._token_endpoint
            raise Exception(msg)

        self._token_endpoint = resp_json["token_endpoint"]
        self._token_endpoint_resolved = time.time()
        self.log.info(f"Resolved token endpoint of '{self._token_endpoint}' for issuer {self.token_issuer}")
        return self._token_endpoint


    def refresh_access_token(self, username, token_name):
        """
        Execute the client credentials flow with the remote token endpoint.

        If successful, write the access token to disk.
        """

        info = self.generate_access_token_info(username, token_name)

        scopes = info.scopes
        if info.profile == "wlcg" or info.profile == "wlcg:1.0":
            scopes.append("wlcg")
        scopes = " ".join(scopes)
        if info.sub:
            scopes.append(f"condor.user:{urllib.parse.quote(info.sub)}")

        fields = {
            "client_id": self.client_id,
            "scopes": scopes,
            "grant_type": "client_credentials",
        }
        if info.audience:
            fields["audience"] = " ".join(info.audience)

        self.log.debug(f"Requesting token from {self.token_endpoint} with following info: {fields}")
        fields["client_secret"] = self.secret

        resp = urllib3.request("POST", self.token_endpoint, fields=fields)
        del fields["client_secret"]
        if resp.status != 200:
            self.log.error(f"HTTP status failure ({resp.status}) when requesting token from {self.token_endpoint} with parameters {fields}")
            return False

        try:
            resp_json = resp.json()
        except Exception as exc:
            self.log.error(f"Token issuer {self.token_issuer} failed to have a valid JSON response: {exc}")
            return False

        access_token = resp_json.get("access_token")
        if not access_token:
            self.log.error(f"Token issuer {self.token_issuer} failed to respond with an access token")
            return False

        lifetime = resp_json.get("expires_in")
        if not lifetime:
            self.log.warning(f"Token issuer {self.token_issuer} failed to indicate when access token will expire; assuming {self.token_lifetime}")

        return self.write_access_token(username, token_name, lifetime, access_token)




#!/usr/bin/env pytest

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import htcondor2
from htcondor2 import (
    JobEventType,
)

from ornithology import (
    action,
    Condor,
    ClusterState,
    DaemonLog,
    JobStatus,
)


#
# Tests to see if the starter is actually correctly executing the guidance
# it's given.  We could add more-specific testing to see if the diagnostic
# tool(s) are working as necessary (probably as a lambda in the test case).
#


TEST_CASES = {
    "CarryOn": (
        '{ [ Command = "CarryOn"; ] }',
        JobStatus.HELD,
        "Carrying on according to guidance...",
    ),
    "StartJob": (
        '{ [ Command = "StartJob"; ] }',
        JobStatus.COMPLETED,
        "Starting job as guided...",
    ),
    "Abort": (
        '{ [ Command = "Abort"; ] }',
        JobStatus.HELD,
        "Aborting job as guided...",
    ),
    "RetryTransfer": (
        '{ [ Command = "RetryTransfer"; ], [ Command = "RetryRequest"; RetryDelay = 5; ], [ Command = "CarryOn"; ] }',
        JobStatus.HELD,
        "Retrying transfer as guided...",
    ),
    "RunDiagnostic": (
        '{ [ Command = "RunDiagnostic"; Diagnostic = "send_ep_logs"; ], [ Command = "CarryOn"; ] }',
        JobStatus.HELD,
        "Running diagnostic 'send_ep_logs' as guided...",
    ),
    # We check the starter side of this test case in test_starter_guidance;
    # the goal here is to verify that the shadow handle the reply correctly.
    "RunUnknownDiagnostic": (
        '{ [ Command = "RunDiagnostic"; Diagnostic = "unknown"; ], [ Command = "CarryOn"; ] }',
        JobStatus.HELD,
        "... diagnostic 'unknown' not registered in param table",
    ),
    "CarryOn w/ Extra": (
        '{ [ Command = "CarryOn"; Extraneous = True; ] }',
        JobStatus.HELD,
        "Carrying on according to guidance...",
    ),
    "StartJob w/ Extra": (
        '{ [ Command = "StartJob"; Extraneous = True; ] }',
        JobStatus.COMPLETED,
        "Starting job as guided...",
    ),
    "Abort w/ Extra": (
        '{ [ Command = "Abort"; Extraneous = True; ] }',
        JobStatus.HELD,
        "Aborting job as guided...",
    ),
    "RetryTransfer w/ Extra": (
        '{ [ Command = "RetryTransfer"; Extraneous = True; ], [ Command = "RetryRequest"; RetryDelay = 5; Extraneous = False; ], [ Command = "CarryOn"; Extraneous = True; ] }',
        JobStatus.HELD,
        "Retrying transfer as guided...",
    ),
    "RunDiagnostic w/ Extra": (
        '{ [ Command = "RunDiagnostic"; Diagnostic = "send_ep_logs"; Extraneous = True; ], [ Command = "CarryOn"; Extraneous = True; ] }',
        JobStatus.HELD,
        "Running diagnostic 'send_ep_logs' as guided...",
    ),
}


@action
def path_to_shadow_wrapper(test_dir):
    return test_dir / "shadow_wrapper"


@action
def the_condor(test_dir, path_to_shadow_wrapper):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "SHADOW":                       path_to_shadow_wrapper.as_posix(),
            "SHADOW_DEBUG":                 "D_FULLDEBUG",
            "STARTD_ENVIRONMENT":           ";http_proxy=;https_proxy=",

            # For simplicity, so that each test job gets its own starter log.
            "STARTER_LOG_NAME_APPEND":      "JobID",
        },
    ) as the_condor:
        SBIN = htcondor2.param["SBIN"]
        path_to_shadow_wrapper.write_text(
            "#!/bin/bash\n"
            f'exec {SBIN}/condor_shadow --use-guidance-in-job-ad "$@"' "\n"
        )
        path_to_shadow_wrapper.chmod(0o777)

        yield the_condor


@action
def the_job_description(test_dir, path_to_sleep):
    return {
        "executable":               path_to_sleep,
        "transfer_executable":      False,
        "should_transfer_files":    True,
        "universe":                 "vanilla",
        "arguments":                5,
        "log":                      f'{(test_dir / "job.log").as_posix()}.$(CLUSTER)',
        "starter_debug":            "D_FULLDEBUG",
        "request_cpus":             1,
        "request_memory":           1,
        "transfer_input_files":     "http://no-such.tld/example",
    }


@action
def the_job_handles(the_condor, the_job_description):
    job_handles = {}
    for name, test_case in TEST_CASES.items():
        (the_guidance, the_expected, _) = test_case

        complete_job_description = {
            ** the_job_description,
            "+_condor_guidance_test_case": the_guidance,
        }
        job_handle = the_condor.submit(
            description=complete_job_description,
            count=1,
        )

        job_handles[name] = job_handle

    yield job_handles


@action(params={name: name for name in TEST_CASES})
def the_job_tuple(request, the_job_handles):
    return (request.param, the_job_handles[request.param])


@action
def the_job_name(the_job_tuple):
    return the_job_tuple[0]


@action
def the_job_handle(the_job_tuple):
    return the_job_tuple[1]


@action
def the_expected_state(the_job_name):
    return TEST_CASES[the_job_name][1]


@action
def the_expected_log_line(the_job_name):
    return TEST_CASES[the_job_name][2]


@action
def the_completed_job(the_condor, the_job_handle):
    assert the_job_handle.wait(
        timeout=40,
        condition=ClusterState.all_terminal
    )

    return the_job_handle


@action
def the_starter_log(test_dir, the_completed_job):
    starter_log_path = (test_dir / "condor" / "log" / f"StarterLog.{the_completed_job.clusterid}.0")
    starter_log = DaemonLog(starter_log_path)
    return starter_log.open()


# From test_allowed_execute_duration.py
def event_types_in_order(types, events):
    t = 0
    for i in range(0, len(events)):
        event = events[i]

        if event.type == types[t]:
            t += 1
            if t == len(types):
                return True
    else:
        return False


class TestGuidanceCommands:

    def test_guidance_command(self,
        test_dir, the_job_name, the_condor,
        the_completed_job, the_expected_state,
        the_starter_log, the_expected_log_line
    ):
        # Did the job carry through to the expected state?
        assert the_completed_job.state.status_exactly( 1, the_expected_state )

        # Did the start actually execute the guidance it was given?
        assert the_starter_log.wait(
            timeout=10,
            condition=lambda line: the_expected_log_line in line.message,
        )

        # Gross hacks.
        if the_job_name == "RunUnknownDiagnostic":
            # Validate that the shadow handled the error correctly.
            expected_shadow_log_line = "Diagnostic 'unknown' did not complete: 'Error - Unregistered'"
            the_shadow_log = the_condor.shadow_log.open()
            assert the_shadow_log.wait(
                timeout=1,
                condition=lambda line: expected_shadow_log_line in line.message and f"{the_completed_job.clusterid}.0" in line.tags
            )

        if the_job_name.startswith("RetryTransfer"):
            # Validate that the transfer was retried.  (Note the lack of
            # an execute event, indicating that both begin/end pairs of
            # file transfer were input.)
            assert event_types_in_order(
                [
                    JobEventType.SUBMIT,
                    JobEventType.FILE_TRANSFER,
                    JobEventType.FILE_TRANSFER,
                ],
                the_completed_job.event_log.events
            )
            assert event_types_in_order(
                [
                    JobEventType.REMOTE_ERROR,
                    JobEventType.JOB_EVICTED,
                    JobEventType.JOB_HELD,
                ],
                the_completed_job.event_log.events
            )


        # This has been disabled until we figure out if it's a good idea.
        # if the_job_name.startswith("RunDiagnostic"):
            # Validate that the diagnostic was run.
            # diagnostic_log_path = (test_dir / ".diagnostic" / f"send_ep_logs.{the_completed_job.clusterid}.0.0")
            # assert diagnostic_log_path.exists()



#!/usr/bin/env pytest

from ornithology import *

import classad2 as classad


TEST_CASES = {
    "blank": { },
    "none": {
        "TRANSFER": [],
    },
    "all": {
        "INPUT": ["requirements", "TransferInput"],
        "OUTPUT": ["TransferOutput"],
        "CHECKPOINT": ["TransferCheckpoint", "CheckpointExitCode"],
        "TRANSFER": ["requirements"],
    },
    "empty": {
        "INPUT": [],
        "TRANSFER": ["requirements"]
    },
    "spawn": {
        "SPAWN": [ "TestAttr1", "TestAttr2" ],
    }
}


@action(params={name: name for name in TEST_CASES})
def the_test_pair(request):
    return (request.param, TEST_CASES[request.param])


@action
def the_test_name(the_test_pair):
    return the_test_pair[0]


@action
def the_test_case(the_test_pair):
    return the_test_pair[1]


@action
def the_job_epoch_file(test_dir, the_test_name):
    return f"{test_dir}/{the_test_name}/EpochLog"


@action
def the_condor(the_test_name, the_test_case, the_job_epoch_file, test_dir):
    raw_config = f"JOB_EPOCH_HISTORY = {the_job_epoch_file}\n"

    for key, value in the_test_case.items():
        raw_config += f"{key}_JOB_ATTRS = {','.join(value)}\n"

    # print(raw_config)
    with Condor(
        local_dir=test_dir / the_test_name,
        raw_config=raw_config,
    ) as the_condor:
        yield the_condor


@action
def the_completed_job(the_test_name, the_condor, path_to_sleep, test_dir):
    (test_dir / "input.txt").touch()

    handle = the_condor.submit(
        description={
            "executable": path_to_sleep,
            "transfer_executable": False,
            "should_transfer_files": True,
            "log": test_dir / f"{the_test_name}-job.log",
            "transfer_input_files": "null://input.txt",
            "transfer_output_files": "input.txt",
            "output_destination": "null://input.txt",
            "My.TestAttr1": 10,
            "My.TestAttr2": False,
        },
        count=1,
    )

    assert( handle.wait(
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
        timeout=120,
    ))

    return handle


# ROFLcopters!
class EpochLogEntry:

    def __init__(self,
        TYPE, ClusterID, ProcID, RunInstanceID, Owner,
        entry_text
    ):
        self.TYPE = TYPE
        self.ClusterID = ClusterID
        self.ProcID = ProcID
        self.RunInstanceID = RunInstanceID
        self.Owner = Owner
        self.entry_text = entry_text
        self.the_ad = None


    @property
    def ad(self):
        if  self.the_ad is None:
            self.the_ad = classad.parseOne(self.entry_text)
            self.entry_text = None
        return self.the_ad


def readNextEpochLogEntry(f) -> EpochLogEntry:
    # There's almost certainly a more-efficient way to do this.
    buffer =''
    while True:
        line = f.readline()
        if line == "":
            # This changed in Python 3.7, apparently.
            # raise StopIteration("end-of-file")
            return
        elif not line.startswith("***"):
            buffer += line
        else:
            line = line[4:]
            TYPE = line.split()[0]
            line = line[len(f'{TYPE}') + 1:]
            lines = line.replace(" ", "\n")
            headerAd = classad.parseOne(lines)
            yield EpochLogEntry(
                TYPE,
                headerAd['ClusterID'], headerAd['ProcID'],
                headerAd['RunInstanceID'], headerAd['Owner'],
                buffer
            )
            buffer = ''

EXPECTED_ALL_RECORDS = [ "RunInstanceID", "EpochAdType", "EpochWriteDate", "Owner", "ClusterId", "ProcId" ]
EXPECTED_TRANSFER_RECORDS = [ "TransferClass", "RemoteHost" ]

ALWAYS_KEYS = {
    "INPUT": EXPECTED_ALL_RECORDS + EXPECTED_TRANSFER_RECORDS + [ "InputPluginResultList", "InputPluginInvocations" ],
    "OUTPUT": EXPECTED_ALL_RECORDS + EXPECTED_TRANSFER_RECORDS + [ "OutputPluginResultList",  "OutputPluginInvocations" ],
    "CHECKPOINT": EXPECTED_ALL_RECORDS + EXPECTED_TRANSFER_RECORDS + [ "CheckpointPluginResultList", "CheckpointPluginInvocations" ],
    "SPAWN": EXPECTED_ALL_RECORDS + [ "NumShadowStarts", "ShadowBday" ],
}


# We could generate this by interrogating the param table.
DEFAULT_KEYS = {
    "INPUT": [ "NumShadowStarts" ],
    "OUTPUT": [ "NumShadowStarts" ],
    "CHECKPOINT": [ "NumShadowStarts" ],
    "SPAWN": [],
}

class TestEpochAttrs:

    def test_epoch_log(self, the_completed_job, the_job_epoch_file, the_test_case):
        with open(the_job_epoch_file, 'r') as f:
            for entry in readNextEpochLogEntry(f):
                if entry.TYPE == "EPOCH" or entry.TYPE == "STARTER":
                    continue

                expectedKeys = []
                expectedKeys.extend(ALWAYS_KEYS[entry.TYPE])

                typeInfo = the_test_case.get(entry.TYPE)
                if typeInfo is None:
                    defaultTypeInfo = the_test_case.get("TRANSFER") if entry.TYPE != "SPAWN" else None
                    if defaultTypeInfo is None:
                        defaultTypeInfo = DEFAULT_KEYS[entry.TYPE]
                    expectedKeys.extend(defaultTypeInfo)
                else:
                    expectedKeys.extend(typeInfo)

                assert(sorted(entry.ad.keys()) == sorted(expectedKeys))



#!/usr/bin/env pytest

import os
import time

from ornithology import (
    action,
    write_file,
    format_script,
    ClusterState,
    Condor,
)

import htcondor2


import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def condor_files():
    return (
        ".job.ad",
        ".machine.ad",
        ".docker_sock",
        ".docker_stdout",
        ".docker_stderror",
        ".update.ad",
        ".update.ad.tmp",
        ".execution_overlay.ad",
    )


@action
def the_test_script(test_dir):
    test_script_path = test_dir / "test_script.sh"

    script = """
        #!/bin/bash

        if [ ! -f checkpoint-file-1 ]; then
            touch checkpoint-file-1
            exit 85
        fi
        sleep 3600
        touch output-file-1
        exit 0
    """

    write_file(test_script_path, format_script(script))
    return test_script_path


@action
def the_running_test_job(default_condor, the_test_script):
    description = {
        "executable":               str(the_test_script),
        "should_transfer_files":    "yes",
        "transfer_executable":      "yes",
        "log":                      "log",
        "output":                   "output",
        "error":                    "error",
        "checkpoint_exit_code":     85,
    }

    the_test_job = default_condor.submit(
        description = description,
        count = 1
    )

    the_test_job.wait(
        verbose = True,
        timeout = 60,
        condition = ClusterState.all_running,
        fail_condition = ClusterState.any_held,
    )

    yield the_test_job

    the_test_job.remove()


@action
def files_in_spool(default_condor, the_running_test_job):
    clusterID = the_running_test_job.clusterid
    job_spool_dir = default_condor.spool_dir / f"{clusterID}" / "0" / f"cluster{clusterID}.proc0.subproc0"

    list = []
    timeout = 30
    for iteration in range(1, timeout):
        assert(iteration <= timeout)
        time.sleep(1)
        if os.path.exists(job_spool_dir):
            list = os.listdir(job_spool_dir)
            if "checkpoint-file-1" in list:
                return list


class TestUnspecifiedCheckpoint:

    def test_no_condor_files(self, files_in_spool, condor_files):
        for file in condor_files:
            assert file not in files_in_spool



#!/usr/bin/env pytest

import os

from ornithology import (
    action,
    Condor,
)

import htcondor2


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            'CONTAINER_IMAGES_COMMON_BY_DEFAULT': 'FALSE',
            'CONTAINER_REGEX_COMMON_BY_DEFAULT': '',
        }
    ) as the_condor:
        yield the_condor


@action
def the_job_description():
    return """
        universe = vanilla
        shell = cat common-file input-file-$(ProcID) > output-file-$(ProcID); sleep 10
        MY.CommonInputFiles = "common-file"
        transfer_input_files = input-file-$(ProcID)
        transfer_output_files = output-file-$(ProcID)
        should_transfer_files = YES
        requirements = HasCommonFilesTransfer
        request_cpus = 1
        request_memory = 1024
        log = $(ClusterID).log
    """


TEST_CASES = {
    "sif-blank-no": {
        "submit":   "container_image = busybox.sif",
        "regex":    "",
        "expected": False,
    },
    "sif-match-yes": {
        "submit":   "container_image = busybox.sif",
        "regex":    '''^.*\\.sif$''',
        "expected": True,
    },
    "http-match-yes": {
        "submit":    "container_image = http://example.tld/busybox.sif",
        "regex":    '''^http://.*$''',
        "expected": True,
    },
    "osdf-match-yes": {
        "submit":   "container_image = osdf://example/namespace/busybox.sif",
        "regex":    '''^osdf://.*$''',
        "expected": True,
    },
    "sif-other-no": {
        "submit":   "container_image = busybox.sif",
        "regex":    '''^osdf://.*$''',
        "expected": False,
    },
}


@action
def the_cluster_ads(the_condor, the_job_description):
    clusterAds = {}
    schedd = the_condor.get_local_schedd()
    original = htcondor2.param.get('CONTAINER_REGEX_COMMON_BY_DEFAULT', "")

    for name, value in TEST_CASES.items():
        jd = the_job_description + f"{value['submit']}\n"
        htcondor2.param['CONTAINER_REGEX_COMMON_BY_DEFAULT'] = value['regex']
        submit = htcondor2.Submit(jd)
        sr = schedd.submit(submit, count=1)
        clusterAds[name] = sr.clusterad()
    htcondor2.param['CONTAINER_REGEX_COMMON_BY_DEFAULT'] = original

    return clusterAds


class TestCIC_Regex():

    def test_container_commonality_is_as_expected(self, the_cluster_ads):
        for name, ad in the_cluster_ads.items():
            common = ad.get('_x_catalog_condor_container_image') is not None
            print(f"{name} -> {common}")
            assert TEST_CASES[name]['expected'] == common



#!/usr/bin/env pytest

import pytest
import logging

from ornithology import (
    action,
    ClusterState,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_job(default_condor, test_dir):
    the_job = default_condor.submit(
        {
            "executable":           "/bin/bash",
            "arguments":            '''"-c 'exit 17'"''',
            "log":                  (test_dir / "the_job.log").as_posix(),
            "transfer_executable":  "false",
            "leave_in_queue":       "true",
        }
    )

    assert the_job.wait(
        condition=ClusterState.all_terminal,
        timeout=60,
    )

    return the_job


class TestJobExitCode:

    def test_job_exit_code(self, the_job):
        the_job_ad = the_job.query()[0]
        the_exit_code = the_job_ad['ExitCode']
        assert the_exit_code == 17



#!/usr/bin/env pytest

#
# Test that the shadow writes a reconnect record to the ShadowReconnectLog
# CSV file when a reconnect attempt fails because the starter is dead.
#
# Strategy:
#   1. Start a personal condor with SHADOW_LOG_RECONNECT enabled
#   2. Submit a long sleep job with a job lease duration
#   3. Wait for it to start running
#   4. Kill the starter process to break the shadow-starter connection
#   5. The shadow detects the broken connection and attempts to reconnect
#   6. The startd reports the starter is gone, so reconnect fails
#   7. The shadow logs a CSV record to ShadowReconnectLog before exiting
#   8. The schedd requeues the job which runs to completion
#

import os
import re
import signal
import time
import logging

from pathlib import Path

from ornithology import (
    standup,
    action,
    Condor,
    ClusterState,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@standup
def condor(test_dir):
    with Condor(
        local_dir=test_dir / "condor",
        config={
            "SHADOW_LOG_RECONNECT": "true",
            "SHADOW_DEBUG": "D_FULLDEBUG",
            "STARTER_DEBUG": "D_FULLDEBUG",
            "NEGOTIATOR_INTERVAL": "2",
            "STARTER_UPDATE_INTERVAL": "5",
        },
    ) as condor:
        yield condor


@action
def reconnect_log_path(condor):
    """Get the path to the ShadowReconnectLog file."""
    p = condor.run_command(["condor_config_val", "LOG"])
    return Path(p.stdout.strip()) / "ShadowReconnectLog"


@action
def completed_job(condor, test_dir, path_to_sleep):
    """Submit a job, kill its starter to trigger a failed reconnect,
    then let the job be requeued and run to completion."""

    handle = condor.submit(
        description={
            "executable": path_to_sleep,
            "arguments": "5",
            "log": (test_dir / "job.log").as_posix(),
            "+JobLeaseDuration": "40",
        },
    )

    # Wait for the job to start running
    assert handle.wait(
        condition=ClusterState.all_running,
        timeout=60,
        verbose=True,
    )

    # Give the shadow time to establish contact and do a lease renewal
    time.sleep(3)

    # Find the starter PID from the StarterLog for this personal condor.
    # The starter log has a slot suffix (e.g. StarterLog.slot1_1),
    # so glob for all StarterLog* files in the log directory.
    # DaemonCore logs "** PID = <pid>" at startup.
    result = condor.run_command(["condor_config_val", "LOG"])
    log_dir = Path(result.stdout.strip())
    starter_logs = sorted(log_dir.glob("StarterLog.slot*"))
    assert len(starter_logs) > 0, f"No StarterLog.slot* files found in {log_dir}"
    starter_pid = None
    for slog in starter_logs:
        text = slog.read_text()
        pid_matches = re.findall(r"\*\* PID = (\d+)", text)
        if pid_matches:
            starter_pid = int(pid_matches[-1])
            logger.info(f"Found starter PID {starter_pid} from {slog}")
            break
    assert starter_pid is not None, "Could not find starter PID in any StarterLog"

    # Kill the starter to break the shadow-starter connection.
    # The shadow will detect the broken socket, attempt to reconnect,
    # and the startd will report the starter is gone (CA_FAILURE),
    # triggering reconnectFailed() with starter_known_dead=true.
    os.kill(starter_pid, signal.SIGKILL)

    # The job should be requeued and eventually re-run to completion.
    assert handle.wait(
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
        timeout=90,
        verbose=True,
    )

    return handle


class TestShadowReconnectLog:

    def test_reconnect_log_exists(self, completed_job, reconnect_log_path):
        """Verify the ShadowReconnectLog file was created."""
        assert reconnect_log_path.exists(), \
            f"ShadowReconnectLog not found at {reconnect_log_path}"

    def test_reconnect_log_has_record(self, completed_job, reconnect_log_path):
        """Verify the log contains a valid CSV failed-reconnect record."""
        content = reconnect_log_path.read_text()
        lines = [l.strip() for l in content.splitlines() if l.strip()]
        logger.info(f"ShadowReconnectLog contents:\n{content}")

        assert len(lines) >= 1, "ShadowReconnectLog should have at least one record"

        # Parse the CSV line: ID,T1,T2,B1,T3,T4,ID1,B2
        fields = lines[0].split(",")
        assert len(fields) == 8, f"Expected 8 CSV fields, got {len(fields)}: {lines[0]}"

        job_id = fields[0]   # cluster.proc
        t1 = int(fields[1])  # activation time
        t2 = int(fields[2])  # last contact time
        b1 = fields[3]       # reconnect succeeded
        t3 = int(fields[4])  # reconnect time
        t4 = int(fields[5])  # lease duration
        # fields[6] is ID1 (timeout version, may be empty)
        b2 = fields[7]       # starter known dead

        # T1 should be a recent epoch time (within last 2 minutes)
        now = int(time.time())
        assert now - t1 < 120, f"T1 ({t1}) should be recent"

        # T2 should be >= T1
        assert t2 >= t1, f"T2 ({t2}) should be >= T1 ({t1})"

        # B1 should be false (reconnect failed because starter was killed)
        assert b1 == "false", f"B1 should be 'false', got '{b1}'"

        # T4 should be the job lease duration we set (40)
        assert t4 == 40, f"T4 should be 40, got {t4}"

        # B2 should be true (starter is known to be dead)
        assert b2 == "true", f"B2 should be 'true', got '{b2}'"



import sys
import enum


class DaemonCommand(enum.IntEnum):
    """
    An enumeration of daemon commands.

    .. attribute:: DaemonOn
    .. attribute:: DaemonOff
    .. attribute:: DaemonOffFast
    .. attribute:: DaemonOffPeaceful
    .. attribute:: DaemonsOn
    .. attribute:: DaemonsOff
    .. attribute:: DaemonsOffFast
    .. attribute:: DaemonsOffPeaceful
    .. attribute:: OffFast
    .. attribute:: OffForce
    .. attribute:: OffGraceful
    .. attribute:: OffPeaceful
    .. attribute:: Reconfig
    .. attribute:: Restart
    .. attribute:: RestartPeaceful
    .. attribute:: SetForceShutdown
    .. attribute:: SetPeacefulShutdown
    """

    DaemonOn                = 469
    DaemonOff               = 467
    DaemonOffFast           = 468
    DaemonOffPeaceful       = 469
    DaemonsOn               = 483
    DaemonsOff              = 454
    DaemonsOffFast          = 461
    DaemonsOffPeaceful      = 484
    OffFast                 = 60006
    OffForce                = 60042
    OffGraceful             = 60005
    OffPeaceful             = 60015
    Reconfig                = 60004
    Restart                 = 453
    RestartPeaceful         = 485
    SetForceShutdown        = 60041
    SetPeacefulShutdown     = 60016



import time
from typing import Optional

from .htcondor2_impl import _handle as handle_t

from ._job_event import JobEvent

from .htcondor2_impl import (
    _job_event_log_init,
    _job_event_log_next,
    _job_event_log_close,
    _job_event_log_get_offset,
    _job_event_log_set_offset,
)

class JobEventLog():
    """
    Produce an iterator over job events in a user job event log.  You may
    block or poll for new events.

    A pickled :class:`JobEventLog` resumes generating events if and only if,
    after being unpickled, the job event log file is identical except for
    appended events.
    """

    def __init__(self, filename : str):
        """
        :param filename:  A file containing a job event log.
        """
        self._deadline = 0
        self._filename = filename
        self._handle = handle_t()

        if not isinstance(filename, str):
            raise TypeError("filename must be a string")
        if len(filename) == 0:
            raise ValueError("filename must not be empty")
        _job_event_log_init(self, self._handle, filename)


    def events(self, stop_after : Optional[int] = None) -> 'JobEventLog':
        """
        Returns an iterator over events in the log.

        :param stop_after:  After how many seconds should the iterator
            stop waiting for events?

            If :py:obj:`None`, the iterator waits forever (blocks).

            If ``0``, the iterator never waits (does not block;
            a pure polling iterator).

            For any other value, wait (block) for that many seconds for
            a new event, raising :class:`StopIteration` if one does not
            appear.  (This does not invalidate the iterator.)
        """
        if stop_after is None:
            self._deadline = 0
        elif isinstance(stop_after, int):
            self._deadline = int(time.time()) + stop_after
        else:
            # This was HTCondorTypeError in version 1.
            raise TypeError("deadline must be an integer")

        return self


    def close(self) -> None:
        """
        Close any open underlying file.  This object's iterators will no
        longer produce new events.
        """
        return _job_event_log_close(self, self._handle)


    def __iter__(self) -> 'JobEventLog':
        return self


    def __next__(self) -> JobEvent:
        event_text, data =  _job_event_log_next(self, self._handle, self._deadline)
        return JobEvent(data, event_text)


    def __enter__(self):
        self._deadline = 0
        return self


    def __exit__(self, exceptionType, exceptionValue, traceback):
        self.close()
        return False

    #
    # Pickling.
    #

    def __getstate__(self):
        offset = _job_event_log_get_offset(self, self._handle)
        state = self.__dict__.copy()
        del state['_handle']
        return (state, offset)


    def __setstate__(self, t):
        self.__dict__ = t[0]
        self._handle = handle_t()
        offset = t[1]
        _job_event_log_init(self, self._handle, self._filename)
        _job_event_log_set_offset(self, self._handle, offset)



#!/usr/bin/env pytest

#-----------------------------------------------------------------------------------------------
#   test_dagman_service_nodes_lifetime.py
#
#   This test makes sure that service nodes live for the entire
#   life of DAGMan such that a user removing the DAG does not
#   remove the service node jobs when a FINAL node exists.
#
#   Author: Cole Bollig - 2025/03/07
#-----------------------------------------------------------------------------------------------

from ornithology import *
import htcondor2
import os
from time import time as now

#-----------------------------------------------------------------------------------------------
TIMEOUT = 60

NODE_NAME_SERVICE = "S"
NODE_NAME_WORKER  = "W"
NODE_NAME_FINAL   = "F"
DAG_FILE_NAME     = "test.dag"

#-----------------------------------------------------------------------------------------------
@action
def check_queue_script(test_dir):
    """Create script that checks for only the Service node job to be in the queue"""
    SCRIPT = test_dir / "check_queue.py"
    with open(SCRIPT, "w") as f:
        f.write(f"""#!/usr/bin/env python3
import os
import sys
import htcondor2

print("Checking arguments...")

if len(sys.argv) != 2:
    print(f"ERROR: Invalid number of args: {{len(sys.argv)}}")
    sys.exit(1)

dag_id = int(sys.argv[1])
print(f"DAG ID: {{dag_id}}")

print("Query Schedd for placed DAGMan jobs...")

schedd = htcondor2.Schedd()
ads = schedd.query(constraint=f"DAGManJobId=={{dag_id}}", projection=['DagNodeName', 'JobStatus'])

found = False

for ad in ads:
    if ad['DagNodeName'] != '{NODE_NAME_SERVICE}':
        print(f"ERROR: Unexpected node still in queue: {{ad['DagNodeName']}}")
        sys.exit(2)
    elif found:
        print("ERROR: Service node found twice!!!")
        sys.exit(3)
    elif int(ad['JobStatus']) not in [1,2]:
        print("ERROR: Service node not idle or running")
        sys.exit(4)
    else:
        found = True

if not found:
    print("ERROR: Service node not found")
    sys.exit(5)

print("Check successful")

sys.exit(0)
""")
    os.chmod(str(SCRIPT), 0o0777)
    return SCRIPT

#-----------------------------------------------------------------------------------------------
@action
def submit_dag(default_condor, path_to_sleep, check_queue_script, test_dir):
    DAG_CONTENTS = f"""
    SUBMIT-DESCRIPTION sleep {{
        executable  = {path_to_sleep}
        arguments   = $(WAIT_T)
        universe    = local

        WAIT_T = 600

        queue
    }}

    SERVICE {NODE_NAME_SERVICE} sleep
    JOB {NODE_NAME_WORKER} sleep
    FINAL {NODE_NAME_FINAL} sleep

    VARS {NODE_NAME_FINAL} APPEND WAIT_T="3"
    SCRIPT DEBUG script.debug ALL POST {NODE_NAME_FINAL} {check_queue_script} $DAGID
    """

    # Create test dag
    DAG_FILE = test_dir / DAG_FILE_NAME
    with open(DAG_FILE, "w") as f:
        f.write(DAG_CONTENTS)

    DAG = htcondor2.Submit.from_dag(str(DAG_FILE))
    handle = default_condor.submit(DAG)

    # Wait for DAG to begin executing
    assert handle.wait(condition=ClusterState.all_running, timeout=TIMEOUT)

    # Wait for Worker not to have been submitted to remove DAG node scheduler job
    NODES_LOG = test_dir / f"{DAG_FILE}.nodes.log"
    START_T = now()
    JEL = None

    # Keep trying to open event log
    while now() - START_T < TIMEOUT:
        try:
            JEL = htcondor2.JobEventLog(str(NODES_LOG))
            break
        except htcondor2.HTCondorException:
            JEL = None

    assert JEL != None

    START_T = now()
    REMOVE_CALLED = False

    while not REMOVE_CALLED and now() - START_T < TIMEOUT:
        for event in JEL.events(stop_after=1):
            if event.type == htcondor2.JobEventType.SUBMIT:
                node = event["LogNotes"].split()[2]
                if node == NODE_NAME_WORKER:
                    default_condor.run_command(["condor_rm", handle.clusterid])
                    REMOVE_CALLED = True
                    break

    assert REMOVE_CALLED

    return (handle, NODES_LOG)

#-----------------------------------------------------------------------------------------------
@action
def dag_handle(submit_dag):
    return submit_dag[0]

@action
def nodes_log(submit_dag):
    return str(submit_dag[1])

#-----------------------------------------------------------------------------------------------
class NodeInfo():
    def __init__(self, name: str):
        self.name = name
        self.executed = name == NODE_NAME_WORKER
        self.exited = False

    def __bool__(self):
        return self.executed and self.exited

#===============================================================================================
class TestDAGManServieNodeLifetime:
    def test_service_node_lifetime(self, dag_handle):
        """Wait for removal of DAG to be complete (i.e. SERVICE node actually removed)"""
        assert dag_handle.wait(condition=ClusterState.all_terminal, timeout=TIMEOUT)

    def test_node_states(self, nodes_log):
        """Check all expected node states and event logs"""
        NODE_INFO = dict()
        START_T = now()

        JEL = htcondor2.JobEventLog(nodes_log)
        while True:
            assert now() - START_T < TIMEOUT

            for event in JEL.events(stop_after=1):
                if event.type == htcondor2.JobEventType.SUBMIT:
                    node = event["LogNotes"].split()[2]
                    NODE_INFO[event.cluster] = NodeInfo(node)
                elif event.type == htcondor2.JobEventType.EXECUTE:
                    assert event.cluster in NODE_INFO
                    NODE_INFO[event.cluster].executed = True
                elif event.type == htcondor2.JobEventType.JOB_ABORTED:
                    assert event.cluster in NODE_INFO
                    assert NODE_INFO[event.cluster].name in [NODE_NAME_SERVICE, NODE_NAME_WORKER]
                    assert "DAG" in event["Reason"]
                    NODE_INFO[event.cluster].exited = True
                elif event.type == htcondor2.JobEventType.JOB_TERMINATED:
                    assert event.cluster in NODE_INFO
                    assert NODE_INFO[event.cluster].name == NODE_NAME_FINAL
                    assert event["ReturnValue"] == 0
                    NODE_INFO[event.cluster].exited = True

            if all(NODE_INFO.values()):
                break





import os

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives

class DummyDirective(Directive):
    """Dummy Directive to use for man pages"""
    optional_arguments = 10
    has_content = True
    option_spec = {
        "caption" : directives.unchanged,
        "align" : directives.unchanged,
    }

    def run(self) -> list:
        return []

def setup(app):
    """Setup sphinx to know of this directive"""
    assert os.getenv("MANPAGES", "False") in ["True", "true", "1"]
    app.add_directive('mermaid', DummyDirective)
